from tune_sklearn.tune_gridsearch import TuneGridSearchCV
from tune_sklearn.tune_search import TuneSearchCV
from tune_sklearn._version import __version__

__all__ = ["TuneGridSearchCV", "TuneSearchCV", "__version__"]



"""Class for cross-validation over distributions of hyperparameters
    -- Anthony Yu and Michael Chau
"""
import logging
import random

from ray.tune.stopper import CombinedStopper
from sklearn.base import clone
import numpy as np
import warnings
import os

from ray import tune
from ray.tune.sample import Domain
from ray.tune.suggest import ConcurrencyLimiter, BasicVariantGenerator

from tune_sklearn.utils import check_is_pipeline, MaximumIterationStopper
from tune_sklearn.tune_basesearch import TuneBaseSearchCV
from tune_sklearn._trainable import _Trainable, _PipelineTrainable
from tune_sklearn.list_searcher import RandomListSearcher
from tune_sklearn.utils import check_error_warm_start

logger = logging.getLogger(__name__)


def _check_distribution(dist, search_optimization):
    # Tune Domain is always good
    if isinstance(dist, Domain):
        return

    if search_optimization == "random":
        if not (isinstance(dist, list) or hasattr(dist, "rvs")):
            raise ValueError("distribution must be a list or scipy "
                             "distribution when using randomized search")
    elif not isinstance(dist, tuple) and not isinstance(dist, list):
        if search_optimization == "bayesian":
            import skopt
            if not isinstance(dist, skopt.space.Dimension):
                raise ValueError("distribution must be a tuple, list, or "
                                 "`skopt.space.Dimension` instance when using "
                                 "bayesian search")
        elif search_optimization == "hyperopt":
            import hyperopt.pyll
            if not isinstance(dist, hyperopt.pyll.base.Apply):
                raise ValueError(
                    "distribution must be a tuple, list, or "
                    "`hyperopt.pyll.base.Apply` instance when using "
                    "hyperopt search")
        elif search_optimization == "optuna":
            import optuna.distributions
            if not isinstance(dist, optuna.distributions.BaseDistribution):
                raise ValueError("distribution must be a tuple, list, or "
                                 "`optuna.distributions.BaseDistribution`"
                                 "instance when using optuna search")
        elif search_optimization == "bohb":
            import ConfigSpace.hyperparameters
            if not isinstance(dist,
                              ConfigSpace.hyperparameters.Hyperparameter):
                raise ValueError(
                    "distribution must be a tuple, list, or "
                    "`ConfigSpace.hyperparameters.Hyperparameter` "
                    "instance when using bohb search")


class TuneSearchCV(TuneBaseSearchCV):
    """Generic, non-grid search on hyper parameters.

    Randomized search is invoked with ``search_optimization`` set to
    ``"random"`` and behaves like scikit-learn's ``RandomizedSearchCV``.

    Bayesian search can be invoked with several values of
    ``search_optimization``.

     - ``"bayesian"``, using https://scikit-optimize.github.io/stable/
     - ``"bohb"``, using HpBandSter - https://github.com/automl/HpBandSter

    Tree-Parzen Estimators search is invoked with ``search_optimization``
    set to ``"hyperopt"``, using HyperOpt - http://hyperopt.github.io/hyperopt

    All types of search aside from Randomized search require parent
    libraries to be installed.

    TuneSearchCV implements a "fit" and a "score" method.
    It also implements "predict", "predict_proba", "decision_function",
    "transform" and "inverse_transform" if they are implemented in the
    estimator used.

    The parameters of the estimator used to apply these methods are optimized
    by cross-validated search over parameter settings.

    In contrast to GridSearchCV, not all parameter values are tried out, but
    rather a fixed number of parameter settings is sampled from the specified
    distributions. The number of parameter settings that are tried is
    given by n_trials.

    Args:
        estimator (`estimator`): This is assumed to implement the
            scikit-learn estimator interface. Either estimator needs to
            provide a ``score`` function, or ``scoring`` must be passed.
        param_distributions (`dict` or `list` or `ConfigurationSpace`): Serves
            as the ``param_distributions`` parameter in scikit-learn's
            ``RandomizedSearchCV`` or as the ``search_space`` parameter in
            ``BayesSearchCV``.
            For randomized search: dictionary with parameters names (string)
            as keys and distributions or lists of parameter settings to try
            for randomized search.
            Distributions must provide a rvs method for sampling (such as
            those from scipy.stats.distributions). Ray Tune search spaces
            are also supported.
            If a list is given, it is sampled uniformly. If a list of dicts is
            given, first a dict is sampled uniformly, and then a parameter is
            sampled using that dict as above.
            For other types of search: dictionary with parameter names (string)
            as keys. Values can be

            - a (lower_bound, upper_bound) tuple (for Real or Integer params)
            - a (lower_bound, upper_bound, "prior") tuple (for Real params)
            - as a list of categories (for Categorical dimensions)
            - Ray Tune search space (eg. ``tune.uniform``)

            ``"bayesian"`` (scikit-optimize) also accepts

            - skopt.space.Dimension instance (Real, Integer or Categorical).

            ``"hyperopt"`` (HyperOpt) also accepts

            - an instance of a hyperopt.pyll.base.Apply object.

            ``"bohb"`` (HpBandSter) also accepts

            - ConfigSpace.hyperparameters.Hyperparameter instance.

            ``"optuna"`` (Optuna) also accepts

            - an instance of a optuna.distributions.BaseDistribution object.

            For ``"bohb"`` (HpBandSter) it is also possible to pass a
            `ConfigSpace.ConfigurationSpace` object instead of dict or a list.

            https://scikit-optimize.github.io/stable/modules/
            classes.html#module-skopt.space.space
        early_stopping (bool, str or :class:`TrialScheduler`, optional): Option
            to stop fitting to a hyperparameter configuration if it performs
            poorly. Possible inputs are:

            - If True, defaults to ASHAScheduler.
            - A string corresponding to the name of a Tune Trial Scheduler
              (i.e., "ASHAScheduler"). To specify parameters of the scheduler,
              pass in a scheduler object instead of a string.
            - Scheduler for executing fit with early stopping. Only a subset
              of schedulers are currently supported. The scheduler will only be
              used if the estimator supports partial fitting
            - If None or False, early stopping will not be used.

            Unless a ``HyperBandForBOHB`` object is passed,
            this parameter is ignored for ``"bohb"``, as it requires
            ``HyperBandForBOHB``.

        n_trials (int): Number of parameter settings that are sampled.
            n_trials trades off runtime vs quality of the solution.
            Defaults to 10.
        scoring (str, callable, list/tuple, dict, or None): A single
            string or a callable to evaluate the predictions on the test set.
            See https://scikit-learn.org/stable/modules/model_evaluation.html
            #scoring-parameter for all options.
            For evaluating multiple metrics, either give a list/tuple of
            (unique) strings or a dict with names as keys and callables as
            values.
            If None, the estimator's score method is used. Defaults to None.
        n_jobs (int): Number of jobs to run in parallel. None or -1 means
            using all processors. Defaults to None. If set to 1, jobs
            will be run using Ray's 'local mode'. This can
            lead to significant speedups if the model takes < 10 seconds
            to fit due to removing inter-process communication overheads.
        refit (bool, str, or `callable`): Refit an estimator using the
            best found parameters on the whole dataset.
            For multiple metric evaluation, this needs to be a string denoting
            the scorer that would be used to find the best parameters for
            refitting the estimator at the end.
            The refitted estimator is made available at the ``best_estimator_``
            attribute and permits using ``predict`` directly on this
            ``GridSearchCV`` instance.
            Also for multiple metric evaluation, the attributes
            ``best_index_``, ``best_score_`` and ``best_params_`` will only be
            available if ``refit`` is set and all of them will be determined
            w.r.t this specific scorer. If refit not needed, set to False.
            See ``scoring`` parameter to know more about multiple metric
            evaluation. Defaults to True.
        cv (int, `cross-validation generator` or `iterable`): Determines
            the cross-validation splitting strategy.
            Possible inputs for cv are:

            - None, to use the default 5-fold cross validation,
            - integer, to specify the number of folds in a `(Stratified)KFold`,
            - An iterable yielding (train, test) splits as arrays of indices.

            For integer/None inputs, if the estimator is a classifier and ``y``
            is either binary or multiclass, :class:`StratifiedKFold` is used.
            In all other cases, :class:`KFold` is used. Defaults to None.
        verbose (int): Controls the verbosity: 0 = silent, 1 = only status
            updates, 2 = status and trial results. Defaults to 0.
        random_state (int or `RandomState`): Pseudo random number generator
            state used for random uniform
            sampling from lists of possible values instead of scipy.stats
            distributions.
            If int, random_state is the seed used by the random number
            generator;
            If RandomState instance, a seed is sampled from random_state;
            If None, the random number generator is the RandomState instance
            used by np.random and no seed is provided. Defaults to None.
            Ignored when using BOHB.
        error_score ('raise' or int or float): Value to assign to the score if
            an error occurs in estimator
            fitting. If set to 'raise', the error is raised. If a numeric value
            is given, FitFailedWarning is raised. This parameter does not
            affect the refit step, which will always raise the error.
            Defaults to np.nan.
        return_train_score (bool): If ``False``, the ``cv_results_``
            attribute will not include training scores. Defaults to False.
            Computing training scores is used to get insights on how different
            parameter settings impact the overfitting/underfitting trade-off.
            However computing the scores on the training set can be
            computationally expensive and is not strictly required to select
            the parameters that yield the best generalization performance.
        local_dir (str): A string that defines where checkpoints and logs will
            be stored. Defaults to "~/ray_results"
        max_iters (int): Indicates the maximum number of epochs to run for each
            hyperparameter configuration sampled (specified by ``n_trials``).
            This parameter is used for early stopping. Defaults to 1.
            Depending on the classifier type provided, a resource parameter
            (`resource_param = max_iter or n_estimators`) will be detected.
            The value of `resource_param` will be treated as a
            "max resource value", and all classifiers will be
            initialized with `max resource value // max_iters`, where
            `max_iters` is this defined parameter. On each epoch,
            resource_param (max_iter or n_estimators) is
            incremented by `max resource value // max_iters`.
        search_optimization ("random" or "bayesian" or "bohb" or "hyperopt"
            or "optuna"): Randomized search is invoked with
            ``search_optimization`` set to ``"random"`` and behaves like
            scikit-learn's ``RandomizedSearchCV``.

            Bayesian search can be invoked with several values of
            ``search_optimization``.

            - ``"bayesian"`` via https://scikit-optimize.github.io/stable/
            - ``"bohb"`` via http://github.com/automl/HpBandSter

            Tree-Parzen Estimators search is invoked with
            ``search_optimization`` set to ``"hyperopt"`` via HyperOpt:
            http://hyperopt.github.io/hyperopt

            All types of search aside from Randomized search require parent
            libraries to be installed.
        use_gpu (bool): Indicates whether to use gpu for fitting.
            Defaults to False. If True, training will start processes
            with the proper CUDA VISIBLE DEVICE settings set. If a Ray
            cluster has been initialized, all available GPUs will
            be used.
        loggers (list): A list of the names of the Tune loggers as strings
            to be used to log results. Possible values are "tensorboard",
            "csv", "mlflow", and "json"
        pipeline_auto_early_stop (bool): Only relevant if estimator is Pipeline
            object and early_stopping is enabled/True. If True, early stopping
            will be performed on the last stage of the pipeline (which must
            support early stopping). If False, early stopping will be
            determined by 'Pipeline.warm_start' or 'Pipeline.partial_fit'
            capabilities, which are by default not supported by standard
            SKlearn. Defaults to True.
        stopper (ray.tune.stopper.Stopper): Stopper objects passed to
            ``tune.run()``.
        time_budget_s (int|float|datetime.timedelta): Global time budget in
            seconds after which all trials are stopped. Can also be a
            ``datetime.timedelta`` object. The stopping condition is checked
            after receiving a result, i.e. after each training iteration.
        **search_kwargs (Any):
            Additional arguments to pass to the SearchAlgorithms (tune.suggest)
            objects.

    """

    def __init__(self,
                 estimator,
                 param_distributions,
                 early_stopping=None,
                 n_trials=10,
                 scoring=None,
                 n_jobs=None,
                 refit=True,
                 cv=None,
                 verbose=0,
                 random_state=None,
                 error_score=np.nan,
                 return_train_score=False,
                 local_dir="~/ray_results",
                 max_iters=1,
                 search_optimization="random",
                 use_gpu=False,
                 loggers=None,
                 pipeline_auto_early_stop=True,
                 stopper=None,
                 time_budget_s=None,
                 sk_n_jobs=None,
                 **search_kwargs):
        if sk_n_jobs is not None:
            raise ValueError(
                "Tune-sklearn no longer supports nested parallelism "
                "with new versions of joblib/sklearn. Don't set 'sk_n_jobs'.")
        search_optimization = search_optimization.lower()
        available_optimizations = [
            "random",
            "bayesian",  # scikit-optimize/SkOpt
            "bohb",
            "hyperopt",
            "optuna",
        ]
        if (search_optimization not in available_optimizations):
            raise ValueError("Search optimization must be one of "
                             f"{', '.join(available_optimizations)}")

        self._try_import_required_libraries(search_optimization)

        if isinstance(param_distributions, list):
            if search_optimization != "random":
                raise ValueError("list of dictionaries for parameters "
                                 "is not supported for non-random search")

        if isinstance(param_distributions, dict):
            check_param_distributions = [param_distributions]
        else:
            check_param_distributions = param_distributions

        can_use_param_distributions = False

        from ray.tune.schedulers import HyperBandForBOHB
        if search_optimization == "bohb":
            import ConfigSpace as CS
            can_use_param_distributions = isinstance(check_param_distributions,
                                                     CS.ConfigurationSpace)
            if early_stopping is False:
                raise ValueError(
                    "early_stopping must not be False when using BOHB")
            elif not isinstance(early_stopping, HyperBandForBOHB):
                if early_stopping != "HyperBandForBOHB":
                    warnings.warn("Ignoring early_stopping value, "
                                  "as BOHB requires HyperBandForBOHB "
                                  "as the EarlyStopping scheduler")
                early_stopping = "HyperBandForBOHB"
        elif early_stopping == "HyperBandForBOHB" or isinstance(
                early_stopping, HyperBandForBOHB):
            raise ValueError("search_optimization must be set to 'BOHB' "
                             "if early_stopping is set to HyperBandForBOHB")

        if not can_use_param_distributions:
            for p in check_param_distributions:
                for dist in p.values():
                    _check_distribution(dist, search_optimization)

        super(TuneSearchCV, self).__init__(
            estimator=estimator,
            early_stopping=early_stopping,
            scoring=scoring,
            n_jobs=n_jobs or -1,
            cv=cv,
            verbose=verbose,
            refit=refit,
            error_score=error_score,
            return_train_score=return_train_score,
            local_dir=local_dir,
            max_iters=max_iters,
            use_gpu=use_gpu,
            loggers=loggers,
            pipeline_auto_early_stop=pipeline_auto_early_stop,
            stopper=stopper,
            time_budget_s=time_budget_s)

        check_error_warm_start(self.early_stop_type, param_distributions,
                               estimator)

        self.param_distributions = param_distributions
        self.n_trials = n_trials

        self.random_state = random_state
        if isinstance(random_state, np.random.RandomState):
            # For compatibility with all search algorithms we just
            # sample a seed from the random state
            self.seed = random_state.randint(2**32 - 1)
        else:
            self.seed = random_state

        if search_optimization == "random":
            if search_kwargs:
                raise ValueError("Random search does not support "
                                 f"extra args: {search_kwargs}")
        self.search_optimization = search_optimization
        self.search_kwargs = search_kwargs

    def _fill_config_hyperparam(self, config):
        """Fill in the ``config`` dictionary with the hyperparameters.

        Each distribution in ``self.param_distributions`` must implement
        the ``rvs`` method to generate a random variable. The [0] is
        present to extract the single value out of a list, which is returned
        by ``rvs``.

        Args:
            config (`dict`): dictionary to be filled in as the
                configuration for `tune.run`.

        """
        if self.search_optimization != "random":
            return

        if isinstance(self.param_distributions, list):
            return

        samples = 1
        all_lists = True
        for key, distribution in self.param_distributions.items():
            if isinstance(distribution, Domain):
                config[key] = distribution
            elif isinstance(distribution, list):
                import random

                def get_sample(dist):
                    return lambda spec: dist[random.randint(0, len(dist) - 1)]

                config[key] = tune.sample_from(get_sample(distribution))
                samples *= len(distribution)
            else:
                all_lists = False

                def get_sample(dist):
                    return lambda spec: dist.rvs(1)[0]

                config[key] = tune.sample_from(get_sample(distribution))
        if all_lists:
            self.n_trials = min(self.n_trials, samples)

    def _is_param_distributions_all_tune_domains(self):
        return all(
            isinstance(v, Domain) for k, v in self.param_distributions.items())

    def _get_bohb_config_space(self):
        if self._is_param_distributions_all_tune_domains():
            return self.param_distributions

        import ConfigSpace as CS
        config_space = CS.ConfigurationSpace()

        if isinstance(self.param_distributions, CS.ConfigurationSpace):
            return self.param_distributions

        for param_name, space in self.param_distributions.items():
            prior = "uniform"
            param_name = str(param_name)
            if isinstance(space,
                          tuple) and len(space) >= 2 and len(space) <= 3:
                try:
                    low = float(space[0])
                    high = float(space[1])
                except Exception:
                    raise ValueError(
                        "low and high need to be of type float, "
                        f"are of type {type(low)} and {type(high)}") from None
                if len(space) == 3:
                    prior = space[2]
                    if prior not in ["uniform", "log-uniform"]:
                        raise ValueError(
                            "prior needs to be either "
                            f"'uniform' or 'log-uniform', was {prior}")
                config_space.add_hyperparameter(
                    CS.UniformFloatHyperparameter(
                        name=param_name,
                        lower=low,
                        upper=high,
                        log=prior == "log-uniform"))
            elif isinstance(space, list):
                config_space.add_hyperparameter(
                    CS.CategoricalHyperparameter(
                        name=param_name, choices=space))
            else:
                config_space.add_hyperparameter(space)
        return config_space

    def _get_optuna_params(self):
        from ray.tune.suggest.optuna import param
        config_space = []

        for param_name, space in self.param_distributions.items():
            prior = "uniform"
            param_name = str(param_name)
            if isinstance(space,
                          tuple) and len(space) >= 2 and len(space) <= 3:
                try:
                    low = float(space[0])
                    high = float(space[1])
                except Exception:
                    raise ValueError(
                        "low and high need to be of type float, "
                        f"are of type {type(low)} and {type(high)}") from None
                if len(space) == 3:
                    prior = space[2]
                    if prior not in ["uniform", "log-uniform"]:
                        raise ValueError(
                            "prior needs to be either "
                            f"'uniform' or 'log-uniform', was {prior}")
                if prior == "log-uniform":
                    config_space.append(
                        param.suggest_loguniform(param_name, low, high))
                else:
                    config_space.append(
                        param.suggest_uniform(param_name, low, high))
            elif isinstance(space, list):
                config_space.append(
                    param.suggest_categorical(param_name, space))
            else:
                config_space.append(space)
        return config_space

    def _get_hyperopt_params(self):
        from hyperopt import hp
        config_space = {}

        for param_name, space in self.param_distributions.items():
            prior = "uniform"
            param_name = str(param_name)
            if isinstance(space,
                          tuple) and len(space) >= 2 and len(space) <= 3:
                try:
                    low = float(space[0])
                    high = float(space[1])
                except Exception:
                    raise ValueError(
                        "low and high need to be of type float, "
                        f"are of type {type(low)} and {type(high)}") from None
                if len(space) == 3:
                    prior = space[2]
                    if prior not in ["uniform", "log-uniform"]:
                        raise ValueError("prior needs to be either 'uniform' "
                                         f"or 'log-uniform', was {prior}")
                if prior == "log-uniform":
                    config_space[param_name] = hp.loguniform(
                        param_name, np.log(low), np.log(high))
                else:
                    config_space[param_name] = hp.uniform(
                        param_name, low, high)
            elif isinstance(space, list):
                config_space[param_name] = hp.choice(param_name, space)
            else:
                config_space[param_name] = space
        return config_space

    def _try_import_required_libraries(self, search_optimization):
        if search_optimization == "bayesian":
            try:
                import skopt  # noqa: F401
                from skopt import Optimizer  # noqa: F401
                from ray.tune.suggest.skopt import SkOptSearch  # noqa: F401
            except ImportError:
                raise ImportError(
                    "It appears that scikit-optimize is not installed. "
                    "Do: pip install scikit-optimize") from None
        elif search_optimization == "bohb":
            try:
                from ray.tune.suggest.bohb import TuneBOHB  # noqa: F401
                from ray.tune.schedulers import HyperBandForBOHB  # noqa: F401
                import ConfigSpace as CS  # noqa: F401
            except ImportError:
                raise ImportError(
                    "It appears that either HpBandSter or ConfigSpace "
                    "is not installed. "
                    "Do: pip install hpbandster ConfigSpace") from None
        elif search_optimization == "hyperopt":
            try:
                from ray.tune.suggest.hyperopt import HyperOptSearch  # noqa: F401,E501
                from hyperopt import hp  # noqa: F401
            except ImportError:
                raise ImportError("It appears that hyperopt is not installed. "
                                  "Do: pip install hyperopt") from None
        elif search_optimization == "optuna":
            try:
                from ray.tune.suggest.optuna import OptunaSearch, param  # noqa: F401,E501
                import optuna  # noqa: F401
            except ImportError:
                raise ImportError("It appears that optuna is not installed. "
                                  "Do: pip install optuna") from None

    def _tune_run(self, config, resources_per_trial):
        """Wrapper to call ``tune.run``. Multiple estimators are generated when
        early stopping is possible, whereas a single estimator is
        generated when early stopping is not possible.

        Args:
            config (dict): Configurations such as hyperparameters to run
            ``tune.run`` on.
            resources_per_trial (dict): Resources to use per trial within Ray.
                Accepted keys are `cpu`, `gpu` and custom resources, and values
                are integers specifying the number of each resource to use.

        Returns:
            analysis (`ExperimentAnalysis`): Object returned by
                `tune.run`.

        """
        if self.seed is not None:
            random.seed(self.seed)
            np.random.seed(self.seed)

        trainable = _Trainable
        if self.pipeline_auto_early_stop and check_is_pipeline(
                self.estimator) and self.early_stopping:
            trainable = _PipelineTrainable

        max_iter = self.max_iters
        if self.early_stopping is not None:
            config["estimator_list"] = [
                clone(self.estimator) for _ in range(self.n_splits)
            ]
            if hasattr(self.early_stopping, "_max_t_attr"):
                # we want to delegate stopping to schedulers which
                # support it, but we want it to stop eventually, just in case
                # the solution is to make the stop condition very big
                max_iter = self.max_iters * 10
        else:
            config["estimator_list"] = [self.estimator]

        stopper = MaximumIterationStopper(max_iter=max_iter)
        if self.stopper:
            stopper = CombinedStopper(stopper, self.stopper)

        run_args = dict(
            scheduler=self.early_stopping,
            reuse_actors=True,
            verbose=self.verbose,
            stop=stopper,
            num_samples=self.n_trials,
            config=config,
            fail_fast="raise",
            resources_per_trial=resources_per_trial,
            local_dir=os.path.expanduser(self.local_dir),
            loggers=self.loggers,
            time_budget_s=self.time_budget_s)

        if self.search_optimization == "random":
            if isinstance(self.param_distributions, list):
                search_algo = RandomListSearcher(self.param_distributions)
            else:
                search_algo = BasicVariantGenerator()
            run_args["search_alg"] = search_algo
        else:
            search_space = None
            override_search_space = True
            if self._is_param_distributions_all_tune_domains():
                run_args["config"].update(self.param_distributions)
                override_search_space = False

            search_kwargs = self.search_kwargs.copy()
            search_kwargs.update(metric=self._metric_name, mode="max")

            if self.search_optimization == "bayesian":
                from ray.tune.suggest.skopt import SkOptSearch
                if override_search_space:
                    search_space = self.param_distributions
                search_algo = SkOptSearch(space=search_space, **search_kwargs)
                run_args["search_alg"] = search_algo

            elif self.search_optimization == "bohb":
                from ray.tune.suggest.bohb import TuneBOHB
                if override_search_space:
                    search_space = self._get_bohb_config_space()
                if self.seed:
                    warnings.warn("'seed' is not implemented for BOHB.")
                search_algo = TuneBOHB(space=search_space, **search_kwargs)
                # search_algo = TuneBOHB(
                #     space=search_space, seed=self.seed, **search_kwargs)
                run_args["search_alg"] = search_algo

            elif self.search_optimization == "optuna":
                from ray.tune.suggest.optuna import OptunaSearch
                from optuna.samplers import TPESampler
                sampler = TPESampler(seed=self.seed)
                if override_search_space:
                    search_space = self._get_optuna_params()
                search_algo = OptunaSearch(
                    space=search_space, sampler=sampler, **search_kwargs)
                run_args["search_alg"] = search_algo

            elif self.search_optimization == "hyperopt":
                from ray.tune.suggest.hyperopt import HyperOptSearch
                if override_search_space:
                    search_space = self._get_hyperopt_params()
                search_algo = HyperOptSearch(
                    space=search_space,
                    random_state_seed=self.seed,
                    **search_kwargs)
                run_args["search_alg"] = search_algo

            else:
                # This should not happen as we validate the input before
                # this method. Still, just to be sure, raise an error here.
                raise ValueError(
                    f"Invalid search optimizer: {self.search_optimization}")

        if isinstance(self.n_jobs, int) and self.n_jobs > 0 \
           and not self.search_optimization == "random":
            search_algo = ConcurrencyLimiter(
                search_algo, max_concurrent=self.n_jobs)
            run_args["search_alg"] = search_algo

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="fail_fast='raise' "
                "detected.")
            analysis = tune.run(trainable, **run_args)
        return analysis



def has_xgboost():
    try:
        import xgboost  # noqa: F401
        return True
    except ImportError:
        return False


def is_xgboost_model(clf):
    if not has_xgboost():
        return False
    from xgboost.sklearn import XGBModel  # noqa: F401
    return isinstance(clf, XGBModel)


def has_lightgbm():
    try:
        import lightgbm  # noqa: F401
        return True
    except ImportError:
        return False


def has_required_lightgbm_version():
    """
    lightgbm>=3.0.0 is required for early stopping
    """
    try:
        import lightgbm  # noqa: F401
        version = [int(x) for x in lightgbm.__version__.split(".")]
        return version[0] >= 3
    except ImportError:
        return False


def is_lightgbm_model(clf):
    if not has_lightgbm():
        return False
    from lightgbm.sklearn import LGBMModel  # noqa: F401
    return isinstance(clf, LGBMModel)


def is_lightgbm_model_of_required_version(clf):
    if not has_required_lightgbm_version():
        return False
    from lightgbm.sklearn import LGBMModel  # noqa: F401
    return isinstance(clf, LGBMModel)


def has_catboost():
    try:
        import catboost  # noqa: F401
        return True
    except ImportError:
        return False


def is_catboost_model(clf):
    if not has_catboost():
        return False
    from catboost import CatBoost  # noqa: F401
    return isinstance(clf, CatBoost)



from collections import defaultdict
from typing import Dict

from sklearn.metrics import check_scoring
from sklearn.pipeline import Pipeline
from tune_sklearn._detect_booster import (
    is_xgboost_model, is_lightgbm_model_of_required_version, is_catboost_model)
import numpy as np
from enum import Enum, auto
from collections.abc import Sequence

try:
    from ray.tune.stopper import MaximumIterationStopper
except ImportError:
    from ray.tune.stopper import Stopper

    class MaximumIterationStopper(Stopper):
        def __init__(self, max_iter: int):
            self._max_iter = max_iter
            self._iter = defaultdict(lambda: 0)

        def __call__(self, trial_id: str, result: Dict):
            self._iter[trial_id] += 1
            return self._iter[trial_id] >= self._max_iter

        def stop_all(self):
            return False


class EarlyStopping(Enum):
    PARTIAL_FIT = auto()
    WARM_START_ITER = auto()
    WARM_START_ENSEMBLE = auto()
    XGB = auto()
    LGBM = auto()
    CATBOOST = auto()
    NO_EARLY_STOP = auto()


def check_partial_fit(estimator):
    return callable(getattr(estimator, "partial_fit", None))


def check_is_pipeline(estimator):
    return isinstance(estimator, Pipeline)


def check_warm_start_iter(estimator):
    from sklearn.tree import BaseDecisionTree
    from sklearn.ensemble import BaseEnsemble
    is_not_tree_subclass = not issubclass(type(estimator), BaseDecisionTree)
    is_not_ensemble_subclass = not issubclass(type(estimator), BaseEnsemble)

    return (hasattr(estimator, "warm_start")
            and hasattr(estimator, "max_iter") and is_not_ensemble_subclass
            and is_not_tree_subclass)


def check_warm_start_ensemble(estimator):
    from sklearn.ensemble import BaseEnsemble
    is_ensemble_subclass = issubclass(type(estimator), BaseEnsemble)

    return (hasattr(estimator, "warm_start")
            and hasattr(estimator, "n_estimators") and is_ensemble_subclass)


def check_error_warm_start(early_stop_type, estimator_config, estimator):
    if check_is_pipeline(estimator):
        if (early_stop_type == EarlyStopping.WARM_START_ITER
                and f"{estimator.steps[-1][0]}__max_iter" in estimator_config):
            raise ValueError("tune-sklearn uses `max_iter` to warm "
                             "start, so this parameter can't be "
                             "set when warm start early stopping. ")

        if (early_stop_type == EarlyStopping.WARM_START_ENSEMBLE and
                f"{estimator.steps[-1][0]}__n_estimators" in estimator_config):
            raise ValueError("tune-sklearn uses `n_estimators` to warm "
                             "start, so this parameter can't be "
                             "set when warm start early stopping. ")
    else:
        if (early_stop_type == EarlyStopping.WARM_START_ITER
                and "max_iter" in estimator_config):
            raise ValueError("tune-sklearn uses `max_iter` to warm "
                             "start, so this parameter can't be "
                             "set when warm start early stopping. ")
        if (early_stop_type == EarlyStopping.WARM_START_ENSEMBLE
                and "n_estimators" in estimator_config):
            raise ValueError("tune-sklearn uses `n_estimators` to warm "
                             "start, so this parameter can't be "
                             "set when warm start early stopping. ")


def get_early_stop_type(estimator, early_stopping):
    if not early_stopping:
        return EarlyStopping.NO_EARLY_STOP
    can_partial_fit = check_partial_fit(estimator)
    can_warm_start_iter = check_warm_start_iter(estimator)
    can_warm_start_ensemble = check_warm_start_ensemble(estimator)
    is_xgb = is_xgboost_model(estimator)
    is_lgbm = is_lightgbm_model_of_required_version(estimator)
    is_catboost = is_catboost_model(estimator)
    if is_xgb:
        return EarlyStopping.XGB
    elif is_lgbm:
        return EarlyStopping.LGBM
    elif is_catboost:
        return EarlyStopping.CATBOOST
    elif can_partial_fit:
        return EarlyStopping.PARTIAL_FIT
    elif can_warm_start_iter:
        return EarlyStopping.WARM_START_ITER
    elif can_warm_start_ensemble:
        return EarlyStopping.WARM_START_ENSEMBLE
    else:
        return EarlyStopping.NO_EARLY_STOP


def _aggregate_score_dicts(scores):
    """Aggregate the list of dict to dict of np ndarray
    The aggregated output of _fit_and_score will be a list of dict
    of form [{'prec': 0.1, 'acc':1.0}, {'prec': 0.1, 'acc':1.0}, ...]
    Convert it to a dict of array {'prec': np.array([0.1 ...]), ...}
    Parameters
    ----------
    scores : list of dict
        List of dicts of the scores for all scorers. This is a flat list,
        assumed originally to be of row major order.
    Example
    -------
    >>> scores = [{'a': 1, 'b':10}, {'a': 2, 'b':2}, {'a': 3, 'b':3},
    ...           {'a': 10, 'b': 10}]                         # doctest: +SKIP
    >>> _aggregate_score_dicts(scores)                        # doctest: +SKIP
    {'a': array([1, 2, 3, 10]),
     'b': array([10, 2, 3, 10])}
    """
    return {
        key: np.asarray([score[key] for score in scores])
        for key in scores[0]
    }


def _check_multimetric_scoring(estimator, scoring=None):
    """Check the scoring parameter in cases when multiple metrics are allowed
    Parameters
    ----------
    estimator : sklearn estimator instance
        The estimator for which the scoring will be applied.
    scoring : string, callable, list/tuple, dict or None, default: None
        A single string (see :ref:`scoring_parameter`) or a callable
        (see :ref:`scoring`) to evaluate the predictions on the test set.
        For evaluating multiple metrics, either give a list of (unique) strings
        or a dict with names as keys and callables as values.
        NOTE that when using custom scorers, each scorer should return a single
        value. Metric functions returning a list/array of values can be wrapped
        into multiple scorers that return one value each.
        See :ref:`multimetric_grid_search` for an example.
        If None the estimator's score method is used.
        The return value in that case will be ``{'score': <default_scorer>}``.
        If the estimator's score method is not available, a ``TypeError``
        is raised.
    Returns
    -------
    scorers_dict : dict
        A dict mapping each scorer name to its validated scorer.
    is_multimetric : bool
        True if scorer is a list/tuple or dict of callables
        False if scorer is None/str/callable
    """
    if callable(scoring) or scoring is None or isinstance(scoring, str):
        scorers = {"score": check_scoring(estimator, scoring=scoring)}
        return scorers, False
    else:
        err_msg_generic = (
            "scoring should either be a single string or "
            "callable for single metric evaluation or a "
            "list/tuple of strings or a dict of scorer name "
            "mapped to the callable for multiple metric "
            "evaluation. Got %s of type %s" % (repr(scoring), type(scoring)))

        if isinstance(scoring, (list, tuple, set)):
            err_msg = ("The list/tuple elements must be unique "
                       "strings of predefined scorers. ")
            invalid = False
            try:
                keys = set(scoring)
            except TypeError:
                invalid = True
            if invalid:
                raise ValueError(err_msg)

            if len(keys) != len(scoring):
                raise ValueError(err_msg + "Duplicate elements were found in"
                                 " the given list. %r" % repr(scoring))
            elif len(keys) > 0:
                if not all(isinstance(k, str) for k in keys):
                    if any(callable(k) for k in keys):
                        raise ValueError(err_msg +
                                         "One or more of the elements were "
                                         "callables. Use a dict of score name "
                                         "mapped to the scorer callable. "
                                         "Got %r" % repr(scoring))
                    else:
                        raise ValueError(
                            err_msg + "Non-string types were found in "
                            "the given list. Got %r" % repr(scoring))
                scorers = {
                    scorer: check_scoring(estimator, scoring=scorer)
                    for scorer in scoring
                }
            else:
                raise ValueError(err_msg +
                                 "Empty list was given. %r" % repr(scoring))

        elif isinstance(scoring, dict):
            keys = set(scoring)
            if not all(isinstance(k, str) for k in keys):
                raise ValueError("Non-string types were found in the keys of "
                                 "the given dict. scoring=%r" % repr(scoring))
            if len(keys) == 0:
                raise ValueError(
                    "An empty dict was passed. %r" % repr(scoring))
            scorers = {
                key: check_scoring(estimator, scoring=scorer)
                for key, scorer in scoring.items()
            }
        else:
            raise ValueError(err_msg_generic)
        return scorers, True


def is_tune_grid_search(obj):
    """Checks if obj is a dictionary returned by tune.grid_search.
    Returns bool.
    """
    return isinstance(
        obj, dict) and len(obj) == 1 and "grid_search" in obj and isinstance(
            obj["grid_search"], list)


# adapted from sklearn.model_selection._search
def _check_param_grid_tune_grid_search(param_grid):
    if hasattr(param_grid, "items"):
        param_grid = [param_grid]

    for p in param_grid:
        for name, v in p.items():
            if is_tune_grid_search(v):
                continue

            if isinstance(v, np.ndarray) and v.ndim > 1:
                raise ValueError("Parameter array should be one-dimensional.")

            if (isinstance(v, str) or not isinstance(v,
                                                     (np.ndarray, Sequence))):
                raise ValueError("Parameter grid for parameter ({0}) needs to"
                                 " be a tune.grid_search, list or numpy array,"
                                 " but got ({1})."
                                 " Single values need to be wrapped in a list"
                                 " with one element.".format(name, type(v)))

            if len(v) == 0:
                raise ValueError("Parameter values for parameter ({0}) need "
                                 "to be a non-empty sequence.".format(name))



"""Helper class to support passing a
list of dictionaries for hyperparameters
    -- Anthony Yu and Michael Chau
"""

from sklearn.model_selection import ParameterGrid
from ray.tune.suggest.suggestion import Searcher
import random


class ListSearcher(Searcher):
    """Custom search algorithm to support passing in a list of
    dictionaries to TuneGridSearchCV

    """

    def __init__(self, param_grid):
        self._configurations = list(ParameterGrid(param_grid))
        Searcher.__init__(self)

    def suggest(self, trial_id):
        if self._configurations:
            return self._configurations.pop(0)

    def on_trial_complete(self, **kwargs):
        pass


class RandomListSearcher(Searcher):
    """Custom search algorithm to support passing in a list of
    dictionaries to TuneSearchCV for randomized search

    """

    def __init__(self, param_grid):
        self._configurations = param_grid
        Searcher.__init__(self)

    def suggest(self, trial_id):
        selected_dict = self._configurations[random.randint(
            0,
            len(self._configurations) - 1)]
        generated_config = {}

        for key, distribution in selected_dict.items():
            if isinstance(distribution, list):
                generated_config[key] = distribution[random.randint(
                    0,
                    len(distribution) - 1)]
            else:
                generated_config[key] = distribution.rvs(1)[0]

        return generated_config

    def on_trial_complete(self, **kwargs):
        pass



"""Class for doing grid search over lists of hyperparameters
    -- Anthony Yu and Michael Chau
"""
import warnings
import os

from ray.tune.stopper import CombinedStopper
from sklearn.base import clone
from sklearn.model_selection import ParameterGrid
from ray import tune
from tune_sklearn.list_searcher import ListSearcher
from tune_sklearn.utils import (_check_param_grid_tune_grid_search,
                                check_is_pipeline, check_error_warm_start,
                                is_tune_grid_search, MaximumIterationStopper)
from tune_sklearn.tune_basesearch import TuneBaseSearchCV
from tune_sklearn._trainable import _Trainable
from tune_sklearn._trainable import _PipelineTrainable


class TuneGridSearchCV(TuneBaseSearchCV):
    """Exhaustive search over specified parameter values for an estimator.

    Important members are fit, predict.

    GridSearchCV implements a "fit" and a "score" method. It also implements
    "predict", "predict_proba", "decision_function", "transform" and
    "inverse_transform" if they are implemented in the estimator used.

    The parameters of the estimator used to apply these methods are optimized
    by cross-validated grid-search over a parameter grid.

    Args:
        estimator (`estimator`): Object that implements the
            scikit-learn estimator interface. Either estimator needs to
            provide a ``score`` function, or ``scoring`` must be passed.
        param_grid (`dict` or `list` of `dict`): Dictionary with parameters
            names (string) as keys and lists or tune.grid_search outputs of
            parameter settings to try as values, or a list of such
            dictionaries, in which case the grids spanned by each dictionary
            in the list are explored. This enables searching over any sequence
            of parameter settings.
        early_stopping (bool, str or :class:`TrialScheduler`, optional): Option
            to stop fitting to a hyperparameter configuration if it performs
            poorly. Possible inputs are:

            - If True, defaults to ASHAScheduler.
            - A string corresponding to the name of a Tune Trial Scheduler
              (i.e., "ASHAScheduler"). To specify parameters of the scheduler,
              pass in a scheduler object instead of a string.
            - Scheduler for executing fit with early stopping. Only a subset
              of schedulers are currently supported. The scheduler will only be
              used if the estimator supports partial fitting
            - If None or False, early stopping will not be used.

        scoring (str, list/tuple, dict, or None): A single
            string or a callable to evaluate the predictions on the test set.
            See https://scikit-learn.org/stable/modules/model_evaluation.html
            #scoring-parameter for all options.
            For evaluating multiple metrics, either give a list/tuple of
            (unique) strings or a dict with names as keys and callables as
            values.
            If None, the estimator's score method is used. Defaults to None.
        n_jobs (int): Number of jobs to run in parallel. None or -1 means
            using all processors. Defaults to None. If set to 1, jobs
            will be run using Ray's 'local mode'. This can
            lead to significant speedups if the model takes < 10 seconds
            to fit due to removing inter-process communication overheads.
        cv (int, `cross-validation generator` or `iterable`): Determines the
            cross-validation splitting strategy. Possible inputs for cv are:

            - None, to use the default 5-fold cross validation,
            - integer, to specify the number of folds in a `(Stratified)KFold`,
            - An iterable yielding (train, test) splits as arrays of indices.

            For integer/None inputs, if the estimator is a classifier and ``y``
            is either binary or multiclass, :class:`StratifiedKFold` is used.
            In all other cases, :class:`KFold` is used. Defaults to None.
        refit (bool or str):
            Refit an estimator using the best found parameters on the whole
            dataset.
            For multiple metric evaluation, this needs to be a string denoting
            the scorer that would be used to find the best parameters for
            refitting the estimator at the end.
            The refitted estimator is made available at the ``best_estimator_``
            attribute and permits using ``predict`` directly on this
            ``GridSearchCV`` instance.
            Also for multiple metric evaluation, the attributes
            ``best_index_``, ``best_score_`` and ``best_params_`` will only be
            available if ``refit`` is set and all of them will be determined
            w.r.t this specific scorer. If refit not needed, set to False.
            See ``scoring`` parameter to know more about multiple metric
            evaluation.
            Defaults to True.
        verbose (int): Controls the verbosity: 0 = silent, 1 = only
            status updates, 2 = status and trial results. Defaults to 0.
        error_score ('raise' or int or float): Value to assign to the score
            if an error occurs in estimator fitting. If set to 'raise',
            the error is raised. If a numeric value
            is given, FitFailedWarning is raised. This parameter does not
            affect the refit step, which will always raise the error.
            Defaults to np.nan.
        return_train_score (bool): If ``False``, the ``cv_results_``
            attribute will not include training scores. Defaults to False.
            Computing training scores is used to get insights on how different
            parameter settings impact the overfitting/underfitting trade-off.
            However computing the scores on the training set can be
            computationally expensive and is not strictly required to select
            the parameters that yield the best generalization performance.
        local_dir (str): A string that defines where checkpoints will
            be stored. Defaults to "~/ray_results".
        max_iters (int): Indicates the maximum number of epochs to run for each
            hyperparameter configuration sampled.
            This parameter is used for early stopping. Defaults to 1.
            Depending on the classifier type provided, a resource parameter
            (`resource_param = max_iter or n_estimators`) will be detected.
            The value of `resource_param` will be treated as a
            "max resource value", and all classifiers will be
            initialized with `max resource value // max_iters`, where
            `max_iters` is this defined parameter. On each epoch,
            resource_param (max_iter or n_estimators) is
            incremented by `max resource value // max_iters`.
        use_gpu (bool): Indicates whether to use gpu for fitting.
            Defaults to False. If True, training will start processes
            with the proper CUDA VISIBLE DEVICE settings set. If a Ray
            cluster has been initialized, all available GPUs will
            be used.
        loggers (list): A list of the names of the Tune loggers as strings
            to be used to log results. Possible values are "tensorboard",
            "csv", "mlflow", and "json"
        pipeline_auto_early_stop (bool): Only relevant if estimator is Pipeline
            object and early_stopping is enabled/True. If True, early stopping
            will be performed on the last stage of the pipeline (which must
            support early stopping). If False, early stopping will be
            determined by 'Pipeline.warm_start' or 'Pipeline.partial_fit'
            capabilities, which are by default not supported by standard
            SKlearn. Defaults to True.
        stopper (ray.tune.stopper.Stopper): Stopper objects passed to
            ``tune.run()``.
        time_budget_s (int|float|datetime.timedelta): Global time budget in
            seconds after which all trials are stopped. Can also be a
            ``datetime.timedelta`` object.
    """

    def __init__(self,
                 estimator,
                 param_grid,
                 early_stopping=None,
                 scoring=None,
                 n_jobs=None,
                 cv=5,
                 refit=True,
                 verbose=0,
                 error_score="raise",
                 return_train_score=False,
                 local_dir="~/ray_results",
                 max_iters=1,
                 use_gpu=False,
                 loggers=None,
                 pipeline_auto_early_stop=True,
                 stopper=None,
                 time_budget_s=None,
                 sk_n_jobs=None):
        if sk_n_jobs is not None:
            raise ValueError(
                "Tune-sklearn no longer supports nested parallelism "
                "with new versions of joblib/sklearn. Don't set 'sk_n_jobs'.")
        super(TuneGridSearchCV, self).__init__(
            estimator=estimator,
            early_stopping=early_stopping,
            scoring=scoring,
            n_jobs=n_jobs or -1,
            cv=cv,
            refit=refit,
            error_score=error_score,
            return_train_score=return_train_score,
            local_dir=local_dir,
            max_iters=max_iters,
            verbose=verbose,
            use_gpu=use_gpu,
            loggers=loggers,
            pipeline_auto_early_stop=pipeline_auto_early_stop,
            stopper=stopper,
            time_budget_s=time_budget_s)

        check_error_warm_start(self.early_stop_type, param_grid, estimator)

        _check_param_grid_tune_grid_search(param_grid)
        self.param_grid = param_grid

    def _fill_config_hyperparam(self, config):
        """Fill in the ``config`` dictionary with the hyperparameters.

        Each distribution is converted to a list, then returns a
        dictionary showing the values of the hyperparameters that
        have been grid searched over.

        Args:
            config (`dict`): dictionary to be filled in as the
                configuration for `tune.run`.

        """
        if isinstance(self.param_grid, list):
            return

        for key, distribution in self.param_grid.items():
            if is_tune_grid_search(distribution):
                config[key] = distribution
            else:
                config[key] = tune.grid_search(list(distribution))

    def _list_grid_num_samples(self):
        """Calculate the num_samples for `tune.run`.

        This is used when a list of dictionaries is passed in
        for the `param_grid`
        """
        return len(list(ParameterGrid(self.param_grid)))

    def _tune_run(self, config, resources_per_trial):
        """Wrapper to call ``tune.run``. Multiple estimators are generated when
        early stopping is possible, whereas a single estimator is
        generated when  early stopping is not possible.

        Args:
            config (dict): Configurations such as hyperparameters to run
                ``tune.run`` on.
            resources_per_trial (dict): Resources to use per trial within Ray.
                Accepted keys are `cpu`, `gpu` and custom resources, and values
                are integers specifying the number of each resource to use.

        Returns:
            analysis (`ExperimentAnalysis`): Object returned by
                `tune.run`.

        """
        trainable = _Trainable
        if self.pipeline_auto_early_stop and check_is_pipeline(
                self.estimator) and self.early_stopping:
            trainable = _PipelineTrainable

        if self.early_stopping is not None:
            config["estimator_list"] = [
                clone(self.estimator) for _ in range(self.n_splits)
            ]
        else:
            config["estimator_list"] = [self.estimator]

        stopper = MaximumIterationStopper(max_iter=self.max_iters)
        if self.stopper:
            stopper = CombinedStopper(stopper, self.stopper)

        run_args = dict(
            scheduler=self.early_stopping,
            reuse_actors=True,
            verbose=self.verbose,
            stop=stopper,
            config=config,
            fail_fast="raise",
            resources_per_trial=resources_per_trial,
            local_dir=os.path.expanduser(self.local_dir),
            loggers=self.loggers,
            time_budget_s=self.time_budget_s)

        if isinstance(self.param_grid, list):
            run_args.update(
                dict(
                    search_alg=ListSearcher(self.param_grid),
                    num_samples=self._list_grid_num_samples()))

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="fail_fast='raise' "
                "detected.")
            analysis = tune.run(trainable, **run_args)
        return analysis



""" Helper class to train models using Ray backend
"""
from sklearn.base import clone
from sklearn.model_selection import cross_validate
from sklearn.utils.metaestimators import _safe_split
import numpy as np
import os
from pickle import PicklingError
import warnings
import inspect

import ray
from ray.tune import Trainable
import ray.cloudpickle as cpickle
from tune_sklearn.utils import (EarlyStopping, _aggregate_score_dicts)


class _Trainable(Trainable):
    """Class to be passed in as the first argument of tune.run to train models.

    Overrides Ray Tune's Trainable class to specify the setup, train, save,
    and restore routines.

    """
    estimator_list = None

    @property
    def main_estimator(self):
        return self.estimator_list[0]

    def setup(self, config):
        # forward-compatbility
        self._setup(config)

    def _setup(self, config):
        """Sets up Trainable attributes during initialization.

        Also sets up parameters for the sklearn estimator passed in.

        Args:
            config (dict): contains necessary parameters to complete the `fit`
                routine for the estimator. Also includes parameters for early
                stopping if it is set to true.

        """
        self.estimator_list = clone(config.pop("estimator_list"))
        self.early_stopping = config.pop("early_stopping")
        self.early_stop_type = config.pop("early_stop_type")
        X_id = config.pop("X_id")
        self.X = ray.get(X_id)

        y_id = config.pop("y_id")
        self.y = ray.get(y_id)
        self.groups = config.pop("groups")
        self.fit_params = config.pop("fit_params")
        self.scoring = config.pop("scoring")
        self.max_iters = config.pop("max_iters")
        self.cv = config.pop("cv")
        self.return_train_score = config.pop("return_train_score")
        self.n_jobs = config.pop("n_jobs")
        self.metric_name = config.pop("metric_name", "average_test_score")
        self.estimator_config = config
        self.train_accuracy = None
        self.test_accuracy = None
        self.saved_models = []  # XGBoost specific

        if self.early_stopping:
            n_splits = self._setup_early_stopping()

            for i in range(n_splits):
                self.estimator_list[i].set_params(**self.estimator_config)

            if self.early_stop_type in (EarlyStopping.XGB, EarlyStopping.LGBM,
                                        EarlyStopping.CATBOOST):
                self.saved_models = [None for _ in range(n_splits)]
        else:
            self.main_estimator.set_params(**self.estimator_config)

    def _setup_early_stopping(self):
        n_splits = self.cv.get_n_splits(self.X, self.y)
        self.fold_scores = np.empty(n_splits, dtype=dict)
        self.fold_train_scores = np.empty(n_splits, dtype=dict)
        if self.early_stop_type == EarlyStopping.WARM_START_ITER:
            # max_iter here is different than the max_iters the user sets.
            # max_iter is to make sklearn only fit for max_iter/max_iters
            # epochs, while max_iters (which the user can set) is the usual
            # max number of calls to _trainable.
            self.estimator_config["warm_start"] = True
            self.estimator_config[
                "max_iter"] = self.main_estimator.max_iter // self.max_iters

        elif self.early_stop_type == EarlyStopping.WARM_START_ENSEMBLE:
            # Each additional call on a warm start ensemble only trains
            # new estimators added to the ensemble. We start with 0
            # and add self.resource_step estimators before each call to fit
            # in _train(), training the ensemble incrementally.
            self.resource_step = (
                self.main_estimator.n_estimators // self.max_iters)
            self.estimator_config["warm_start"] = True
            self.estimator_config["n_estimators"] = 0

        return n_splits

    def step(self):
        # forward-compatbility
        return self._train()

    def _early_stopping_partial_fit(self, i, estimator, X_train, y_train):
        """Handles early stopping on estimators that support `partial_fit`.

        """
        if "classes" in inspect.getfullargspec(estimator.partial_fit).args:
            estimator.partial_fit(X_train, y_train, classes=np.unique(self.y))
        else:
            estimator.partial_fit(X_train, y_train)

    def _early_stopping_xgb(self, i, estimator, X_train, y_train):
        """Handles early stopping on XGBoost estimators.

        """
        estimator.fit(X_train, y_train, xgb_model=self.saved_models[i])
        self.saved_models[i] = estimator.get_booster()

    def _early_stopping_lgbm(self, i, estimator, X_train, y_train):
        """Handles early stopping on LightGBM estimators.

        """
        estimator.fit(X_train, y_train, init_model=self.saved_models[i])
        self.saved_models[i] = estimator.booster_

    def _early_stopping_catboost(self, i, estimator, X_train, y_train):
        """Handles early stopping on CatBoost estimators.

        """
        estimator.fit(X_train, y_train, init_model=self.saved_models[i])
        self.saved_models[i] = estimator

    def _early_stopping_iter(self, i, estimator, X_train, y_train):
        """Handles early stopping on estimators supporting `warm_start`.

        """
        estimator.fit(X_train, y_train)

    def _early_stopping_ensemble(self, i, estimator, X_train, y_train):
        """Handles early stopping on ensemble estimators.

        """
        # User will not be able to fine tune the n_estimators
        # parameter using ensemble early stopping
        updated_n_estimators = estimator.get_params(
        )["n_estimators"] + self.resource_step
        estimator.set_params(**{"n_estimators": updated_n_estimators})
        estimator.fit(X_train, y_train)

    def _train(self):
        """Trains one iteration of the model called when ``tune.run`` is called.

        Different routines are run depending on if the ``early_stopping``
        attribute is True or not.

        If ``self.early_stopping`` is not None, each fold is fit with
        `partial_fit`, which stops training the model if the validation
        score is not improving for a particular fold.

        Otherwise, run the full cross-validation procedure.

        In both cases, the average test accuracy is returned over all folds,
        as well as the individual folds' accuracies as a dictionary.

        Returns:
            ret (:obj:`dict): Dictionary of results as a basis for
                ``cv_results_`` for one of the cross-validation interfaces.

        """
        if self.early_stopping:
            for i, (train, test) in enumerate(self.cv.split(self.X, self.y)):
                estimator = self.estimator_list[i]
                X_train, y_train = _safe_split(estimator, self.X, self.y,
                                               train)
                X_test, y_test = _safe_split(
                    estimator, self.X, self.y, test, train_indices=train)
                if self.early_stop_type == EarlyStopping.PARTIAL_FIT:
                    self._early_stopping_partial_fit(i, estimator, X_train,
                                                     y_train)
                elif self.early_stop_type == EarlyStopping.XGB:
                    self._early_stopping_xgb(i, estimator, X_train, y_train)
                elif self.early_stop_type == EarlyStopping.LGBM:
                    self._early_stopping_lgbm(i, estimator, X_train, y_train)
                elif self.early_stop_type == EarlyStopping.CATBOOST:
                    self._early_stopping_catboost(i, estimator, X_train,
                                                  y_train)
                elif self.early_stop_type == EarlyStopping.WARM_START_ITER:
                    self._early_stopping_iter(i, estimator, X_train, y_train)
                elif self.early_stop_type == EarlyStopping.WARM_START_ENSEMBLE:
                    self._early_stopping_ensemble(i, estimator, X_train,
                                                  y_train)
                else:
                    raise RuntimeError(
                        "Early stopping set but model is not: "
                        "xgb model, supports partial fit, or warm-startable.")

                if self.return_train_score:
                    self.fold_train_scores[i] = {
                        name: score(estimator, X_train, y_train)
                        for name, score in self.scoring.items()
                    }
                self.fold_scores[i] = {
                    name: score(estimator, X_test, y_test)
                    for name, score in self.scoring.items()
                }

            ret = {}
            agg_fold_scores = _aggregate_score_dicts(self.fold_scores)
            for name, scores in agg_fold_scores.items():
                total = 0
                for i, score in enumerate(scores):
                    total += score
                    key_str = f"split{i}_test_%s" % name
                    ret[key_str] = score
                self.mean_score = total / len(scores)
                ret["average_test_%s" % name] = self.mean_score

            if self.return_train_score:
                agg_fold_train_scores = _aggregate_score_dicts(
                    self.fold_train_scores)
                for name, scores in agg_fold_train_scores.items():
                    total = 0
                    for i, score in enumerate(scores):
                        total += score
                        key_str = f"split{i}_train_%s" % name
                        ret[key_str] = score
                    self.mean_train_score = total / len(scores)
                    ret["average_train_%s" % name] = self.mean_train_score
        else:
            try:
                scores = cross_validate(
                    self.main_estimator,
                    self.X,
                    self.y,
                    cv=self.cv,
                    n_jobs=self.n_jobs,
                    fit_params=self.fit_params,
                    groups=self.groups,
                    scoring=self.scoring,
                    return_train_score=self.return_train_score,
                    error_score="raise")
            except PicklingError:
                warnings.warn("An error occurred in parallelizing the cross "
                              "validation. Proceeding to cross validate with "
                              "one core.")
                scores = cross_validate(
                    self.main_estimator,
                    self.X,
                    self.y,
                    cv=self.cv,
                    fit_params=self.fit_params,
                    groups=self.groups,
                    scoring=self.scoring,
                    return_train_score=self.return_train_score,
                    error_score="raise")

            ret = {}
            for name in self.scoring:
                for i, score in enumerate(scores["test_%s" % name]):
                    key_str = f"split{i}_test_%s" % name
                    ret[key_str] = score
                self.test_accuracy = sum(scores["test_%s" % name]) / len(
                    scores["test_%s" % name])
                ret["average_test_%s" % name] = self.test_accuracy

            if self.return_train_score:
                for name in self.scoring:
                    for i, score in enumerate(scores["train_%s" % name]):
                        key_str = f"split{i}_train_%s" % name
                        ret[key_str] = score
                    self.train_accuracy = sum(scores["train_%s" % name]) / len(
                        scores["train_%s" % name])
                    ret["average_train_%s" % name] = self.train_accuracy

        ret["objective"] = ret.get(self.metric_name)

        return ret

    def save_checkpoint(self, checkpoint_dir):
        # forward-compatbility
        return self._save(checkpoint_dir)

    def _save(self, checkpoint_dir):
        """Creates a checkpoint in ``checkpoint_dir``, creating a pickle file.

        Args:
            checkpoint_dir (str): file path to store pickle checkpoint.

        Returns:
            path (str): file path to the pickled checkpoint file.

        """
        path = os.path.join(checkpoint_dir, "checkpoint")
        try:
            with open(path, "wb") as f:
                cpickle.dump(self.estimator_list, f)
        except Exception:
            warnings.warn("Unable to save estimator.", category=RuntimeWarning)
        return path

    def load_checkpoint(self, checkpoint):
        # forward-compatbility
        return self._restore(checkpoint)

    def _restore(self, checkpoint):
        """Loads a checkpoint created from `save`.

        Args:
            checkpoint (str): file path to pickled checkpoint file.

        """
        try:
            with open(checkpoint, "rb") as f:
                self.estimator_list = cpickle.load(f)
        except Exception:
            warnings.warn("No estimator restored", category=RuntimeWarning)

    def reset_config(self, new_config):
        self.config = new_config
        self._setup(new_config)
        return True


class _PipelineTrainable(_Trainable):
    """Class to be passed in as the first argument of tune.run to train models.

    Overrides Ray Tune's Trainable class to specify the setup, train, save,
    and restore routines.

    Runs all checks and configures parameters for the last step of the
    Pipeline.

    """

    @property
    def base_estimator_name(self):
        return self.main_estimator.steps[-1][0]

    @property
    def base_estimator(self):
        return self.main_estimator.steps[-1][1]

    def _setup_early_stopping(self):
        n_splits = self.cv.get_n_splits(self.X, self.y)
        self.fold_scores = np.empty(n_splits, dtype=dict)
        self.fold_train_scores = np.empty(n_splits, dtype=dict)
        if self.early_stop_type == EarlyStopping.WARM_START_ITER:
            # max_iter here is different than the max_iters the user sets.
            # max_iter is to make sklearn only fit for max_iter/max_iters
            # epochs, while max_iters (which the user can set) is the usual
            # max number of calls to _trainable.
            self.estimator_config[
                f"{self.base_estimator_name}__warm_start"] = True
            self.estimator_config[f"{self.base_estimator_name}__max_iter"] = (
                self.base_estimator.max_iter // self.max_iters)

        elif self.early_stop_type == EarlyStopping.WARM_START_ENSEMBLE:
            # Each additional call on a warm start ensemble only trains
            # new estimators added to the ensemble. We start with 0
            # and add self.resource_step estimators before each call to fit
            # in _train(), training the ensemble incrementally.
            self.resource_step = (
                self.base_estimator.n_estimators // self.max_iters)
            self.estimator_config[
                f"{self.base_estimator_name}__warm_start"] = True
            self.estimator_config[
                f"{self.base_estimator_name}__n_estimators"] = 0

        return n_splits

    def _early_stopping_partial_fit(self, i, estimator, X_train, y_train):
        """Handles early stopping on estimators that support `partial_fit`.

        """
        if hasattr(estimator, "partial_fit"):
            estimator.partial_fit(X_train, y_train, np.unique(self.y))
        else:
            # Workaround if the pipeline itself doesn't support partial_fit
            # (default sklearn doesn't). We set the final step to
            # "passthrough", so that it doesn't get executed,
            # do fit_transform to get transformed data, and then
            # call partial_fit on just the final step.
            last_step = estimator.steps[-1][1]
            estimator.steps[-1] = (estimator.steps[-1][0], "passthrough")
            X_train_transformed = estimator.fit_transform(X_train, y_train)
            estimator.steps[-1] = (estimator.steps[-1][0], last_step)
            if "classes" in inspect.getfullargspec(
                    estimator.steps[-1][1].partial_fit).args:
                estimator.steps[-1][1].partial_fit(
                    X_train_transformed, y_train, classes=np.unique(self.y))
            else:
                estimator.steps[-1][1].partial_fit(X_train_transformed,
                                                   y_train)

    def _early_stopping_xgb(self, i, estimator, X_train, y_train):
        """Handles early stopping on XGBoost estimators.

        """
        estimator.fit(
            X_train, y_train,
            **{f"{self.base_estimator_name}__xgb_model": self.saved_models[i]})
        self.saved_models[i] = estimator.steps[-1][1].get_booster()

    def _early_stopping_lgbm(self, i, estimator, X_train, y_train):
        """Handles early stopping on LightGBM estimators.

        """
        estimator.fit(
            X_train, y_train, **{
                f"{self.base_estimator_name}__init_model": self.saved_models[i]
            })
        self.saved_models[i] = estimator.steps[-1][1].booster_

    def _early_stopping_catboost(self, i, estimator, X_train, y_train):
        """Handles early stopping on CatBoost estimators.

        """
        estimator.fit(
            X_train, y_train, **{
                f"{self.base_estimator_name}__init_model": self.saved_models[i]
            })
        self.saved_models[i] = estimator.steps[-1][1]

    def _early_stopping_ensemble(self, i, estimator, X_train, y_train):
        """Handles early stopping on ensemble estimators.

        """
        # User will not be able to fine tune the n_estimators
        # parameter using ensemble early stopping
        n_estimator_key = f"{self.base_estimator_name}__n_estimators"
        updated_n_estimators = estimator.get_params(
        )[n_estimator_key] + self.resource_step
        estimator.set_params(**{n_estimator_key: updated_n_estimators})
        estimator.fit(X_train, y_train)



"""Parent class for a cross-validation interface
built with a Ray Tune back-end.

Implementation derived from referencing the equivalent
GridSearchCV interfaces from Dask and Optuna.

https://ray.readthedocs.io/en/latest/tune.html
https://dask.org
https://optuna.org
    -- Anthony Yu and Michael Chau
"""
import logging
from collections import defaultdict

from scipy.stats import rankdata
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted
from sklearn.model_selection import check_cv
from sklearn.base import is_classifier
from sklearn.base import clone
from sklearn.exceptions import NotFittedError

import numpy as np
from numpy.ma import MaskedArray
import pandas as pd
import warnings
import multiprocessing
import os
import inspect
import time
import numbers

import ray
from ray.tune.trial import Trial
from ray.tune.schedulers import (
    PopulationBasedTraining, AsyncHyperBandScheduler, HyperBandScheduler,
    MedianStoppingRule, TrialScheduler, ASHAScheduler, HyperBandForBOHB)
from ray.tune.logger import (TBXLogger, JsonLogger, CSVLogger, MLFLowLogger,
                             Logger)

from tune_sklearn.utils import (EarlyStopping, get_early_stop_type,
                                check_is_pipeline, _check_multimetric_scoring)
from tune_sklearn._detect_booster import is_lightgbm_model

logger = logging.getLogger(__name__)


def resolve_early_stopping(early_stopping, max_iters, metric_name):
    if isinstance(early_stopping, str):
        if early_stopping in TuneBaseSearchCV.defined_schedulers:
            if early_stopping == "PopulationBasedTraining":
                return PopulationBasedTraining(metric=metric_name, mode="max")
            elif early_stopping == "AsyncHyperBandScheduler":
                return AsyncHyperBandScheduler(
                    metric=metric_name, mode="max", max_t=max_iters)
            elif early_stopping == "HyperBandScheduler":
                return HyperBandScheduler(
                    metric=metric_name, mode="max", max_t=max_iters)
            elif early_stopping == "MedianStoppingRule":
                return MedianStoppingRule(metric=metric_name, mode="max")
            elif early_stopping == "ASHAScheduler":
                return ASHAScheduler(
                    metric=metric_name, mode="max", max_t=max_iters)
            elif early_stopping == "HyperBandForBOHB":
                return HyperBandForBOHB(
                    metric=metric_name, mode="max", max_t=max_iters)
        raise ValueError("{} is not a defined scheduler. "
                         "Check the list of available schedulers."
                         .format(early_stopping))
    elif isinstance(early_stopping, TrialScheduler):
        early_stopping._metric = metric_name
        early_stopping._mode = "max"
        return early_stopping
    else:
        raise TypeError("`early_stopping` must be a str, boolean, "
                        f"or tune scheduler. Got {type(early_stopping)}.")


def resolve_loggers(loggers):
    init_loggers = {JsonLogger, CSVLogger}
    if loggers is None:
        return list(init_loggers)

    if not isinstance(loggers, list):
        raise TypeError("`loggers` must be a list of str or tune loggers.")

    for log in loggers:
        if isinstance(log, str):
            if log == "tensorboard":
                init_loggers.add(TBXLogger)
            elif log == "csv":
                init_loggers.add(CSVLogger)
            elif log == "mlflow":
                init_loggers.add(MLFLowLogger)
            elif log == "json":
                init_loggers.add(JsonLogger)
            else:
                raise ValueError(("{} is not one of the defined loggers. " +
                                  str(TuneBaseSearchCV.defined_schedulers))
                                 .format(log))
        elif inspect.isclass(log) and issubclass(log, Logger):
            init_loggers.add(log)
        else:
            raise TypeError("`loggers` must be a list of str or tune "
                            "loggers.")

    return list(init_loggers)


class TuneBaseSearchCV(BaseEstimator):
    """Abstract base class for TuneGridSearchCV and TuneSearchCV"""

    defined_schedulers = [
        "PopulationBasedTraining", "AsyncHyperBandScheduler",
        "HyperBandScheduler", "MedianStoppingRule", "ASHAScheduler",
        "HyperBandForBOHB"
    ]
    defined_loggers = ["tensorboard", "csv", "mlflow", "json"]

    @property
    def _estimator_type(self):
        """str: Returns the estimator's estimator type, such as 'classifier'
        or 'regressor'.

        """
        return self.estimator._estimator_type

    @property
    def best_params_(self):
        """dict: Parameter setting that gave the best results on the hold
        out data.

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

        """
        self._check_is_fitted("best_params_", check_refit="multimetric")
        return self.best_params

    @property
    def best_index_(self):
        """int: The index (of the ``cv_results_`` arrays)
        which corresponds to the best candidate parameter setting.

        The dict at ``search.cv_results_['params'][search.best_index_]`` gives
        the parameter setting for the best model, that gives the highest
        mean score (``search.best_score_``).

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

        """
        self._check_is_fitted("best_index_", check_refit="multimetric")
        return self.best_index

    @property
    def best_score_(self):
        """float: Mean cross-validated score of the best_estimator

        For multi-metric evaluation, this is present only if ``refit``
        is specified.

        """
        self._check_is_fitted("best_score_", check_refit="multimetric")
        return self.best_score

    @property
    def multimetric_(self):
        """bool: Whether evaluation performed was multi-metric."""
        return self.is_multi

    @property
    def classes_(self):
        """list: Get the list of unique classes found in the target `y`."""
        self._check_is_fitted("classes_")
        return self.best_estimator_.classes_

    @property
    def n_splits_(self):
        """int: The number of cross-validation splits (folds/iterations)."""
        self._check_is_fitted("n_splits_", check_refit=False)
        return self.n_splits_

    @property
    def best_estimator_(self):
        """estimator: Estimator that was chosen by the search,
        i.e. estimator which gave highest score (or smallest loss if
        specified) on the left out data. Not available if ``refit=False``.

        See ``refit`` parameter for more information on allowed values.
        """
        self._check_is_fitted("best_estimator_")
        return self.best_estimator

    @property
    def refit_time_(self):
        """float: Seconds used for refitting the best model on the
        whole dataset.

        This is present only if ``refit`` is not False.
        """
        self._check_is_fitted("refit_time_")
        return self.refit_time

    @property
    def scorer_(self):
        """function or a dict: Scorer function used on the held out
        data to choose the best parameters for the model.

        For multi-metric evaluation, this attribute holds the validated
        ``scoring`` dict which maps the scorer key to the scorer callable.
        """
        return self.scoring

    @property
    def decision_function(self):
        """function: Get decision_function on the estimator with the best
        found parameters.

        Only available if ``refit=True`` and the underlying estimator supports
        ``decision_function``.

        """
        self._check_is_fitted("decision_function")
        return self.best_estimator_.decision_function

    @property
    def inverse_transform(self):
        """function: Get inverse_transform on the estimator with the best found
        parameters.

        Only available if the underlying estimator implements
        ``inverse_transform`` and ``refit=True``.

        """
        self._check_is_fitted("inverse_transform")
        return self.best_estimator_.inverse_transform

    @property
    def predict(self):
        """function: Get predict on the estimator with the best found
        parameters.

        Only available if ``refit=True`` and the underlying estimator supports
        ``predict``.

        """
        self._check_is_fitted("predict")
        return self.best_estimator_.predict

    @property
    def predict_log_proba(self):
        """function: Get predict_log_proba on the estimator with the best found
        parameters.

        Only available if ``refit=True`` and the underlying estimator supports
        ``predict_log_proba``.

        """
        self._check_is_fitted("predict_log_proba")
        return self.best_estimator_.predict_log_proba

    @property
    def predict_proba(self):
        """function: Get predict_proba on the estimator with the best found
        parameters.

        Only available if ``refit=True`` and the underlying estimator supports
        ``predict_proba``.

        """
        self._check_is_fitted("predict_proba")
        return self.best_estimator_.predict_proba

    @property
    def transform(self):
        """function: Get transform on the estimator with the best found
        parameters.

        Only available if the underlying estimator supports ``transform`` and
        ``refit=True``.

        """
        self._check_is_fitted("transform")
        return self.best_estimator_.transform

    def _check_params(self):
        """Helper method to see if parameters passed in are valid.

        Raises:
            ValueError: if parameters are invalid.

        """
        if not hasattr(self.estimator, "fit"):
            raise ValueError("estimator must be a scikit-learn estimator.")

    def _check_if_refit(self, attr):
        """Helper method to see if the requested property is available based
        on the `refit` argument.

        Args:
            attr (str): Attribute requested by the user.

        Raises:
            AttributeError: If `self.refit` is False.

        """
        if not self.refit:
            raise AttributeError("'{}' is not a valid attribute with "
                                 "'refit=False'.".format(attr))

    def _check_is_fitted(self, method_name, check_refit=True):
        """Helper method to see if the estimator has been fitted.

        Args:
            method_name (str): String of the method name called from the user.

            check_refit (bool|str): Whether to also check for `self.refit`
                param. If "multimetric", will only check if `self.multimetric`
                param is also True. Defaults to True.

        Raises:
            NotFittedError: If the estimator has not been fitted.
            TypeError: If the estimator is invalid (i.e. doesn't implement
                the sklearn estimator interface).

        """
        if not self.refit:
            if check_refit == "multimetric":
                if self.is_multi:
                    msg = (
                        "This {0} instance was initialized with refit=False. "
                        "For multi-metric evaluation, {1} "
                        "is available only after refitting on the best "
                        "parameters.").format(
                            type(self).__name__, method_name)
                    raise NotFittedError(msg)
            elif check_refit:
                msg = (
                    "This {0} instance was initialized with refit=False. {1} "
                    "is available only after refitting on the best "
                    "parameters.").format(type(self).__name__, method_name)
                raise NotFittedError(msg)
        check_is_fitted(self)

    def _is_multimetric(self, scoring):
        """Helper method to see if multimetric scoring is
        being used

        Args:
            scoring (str, callable, list, tuple, or dict):
                the scoring being used
        """

        return isinstance(scoring, (list, tuple, dict))

    def __init__(self,
                 estimator,
                 early_stopping=None,
                 scoring=None,
                 n_jobs=None,
                 cv=5,
                 refit=True,
                 verbose=0,
                 error_score="raise",
                 return_train_score=False,
                 local_dir="~/ray_results",
                 max_iters=1,
                 use_gpu=False,
                 loggers=None,
                 pipeline_auto_early_stop=True,
                 stopper=None,
                 time_budget_s=None):
        if max_iters < 1:
            raise ValueError("max_iters must be greater than or equal to 1.")
        self.estimator = estimator
        self.base_estimator = estimator
        self.pipeline_auto_early_stop = pipeline_auto_early_stop
        self.stopper = stopper
        self.time_budget_s = time_budget_s

        if self.pipeline_auto_early_stop and check_is_pipeline(estimator):
            _, self.base_estimator = self.base_estimator.steps[-1]

        self.early_stop_type = get_early_stop_type(self.base_estimator,
                                                   bool(early_stopping))

        if not self._can_early_stop():
            if early_stopping:
                raise ValueError("Early stopping is not supported because "
                                 "the estimator does not have `partial_fit`, "
                                 "does not support warm_start, or is a "
                                 "tree classifier. Set "
                                 "`early_stopping=False`.")
        if not early_stopping and max_iters > 1:
            warnings.warn(
                "max_iters is set > 1 but incremental/partial training "
                "is not enabled. To enable partial training, "
                "ensure the estimator has `partial_fit` or "
                "`warm_start` and set `early_stopping=True`. "
                "Automatically setting max_iters=1.",
                category=UserWarning)
            max_iters = 1

        # Get metric scoring name
        self.scoring = scoring
        self.refit = refit
        if not hasattr(self, "is_multi"):
            self.scoring, self.is_multi = _check_multimetric_scoring(
                self.estimator, self.scoring)

        if self.is_multi:
            self._base_metric_name = self.refit
        else:
            self._base_metric_name = "score"

        self._metric_name = "average_test_%s" % self._base_metric_name

        if early_stopping:
            if not self._can_early_stop() and is_lightgbm_model(
                    self.base_estimator):
                warnings.warn("lightgbm>=3.0.0 required for early_stopping "
                              "functionality.")
            assert self._can_early_stop()
            if max_iters == 1:
                warnings.warn(
                    "early_stopping is enabled but max_iters = 1. "
                    "To enable partial training, set max_iters > 1.",
                    category=UserWarning)
            if self.early_stop_type == EarlyStopping.XGB:
                warnings.warn(
                    "tune-sklearn implements incremental learning "
                    "for xgboost models following this: "
                    "https://github.com/dmlc/xgboost/issues/1686. "
                    "This may negatively impact performance. To "
                    "disable, set `early_stopping=False`.",
                    category=UserWarning)
            elif self.early_stop_type == EarlyStopping.LGBM:
                warnings.warn(
                    "tune-sklearn implements incremental learning "
                    "for lightgbm models following this: "
                    "https://lightgbm.readthedocs.io/en/latest/pythonapi/"
                    "lightgbm.LGBMModel.html#lightgbm.LGBMModel.fit "
                    "This may negatively impact performance. To "
                    "disable, set `early_stopping=False`.",
                    category=UserWarning)
            elif self.early_stop_type == EarlyStopping.CATBOOST:
                warnings.warn(
                    "tune-sklearn implements incremental learning "
                    "for Catboost models following this: "
                    "https://catboost.ai/docs/concepts/python-usages-"
                    "examples.html#training-continuation "
                    "This may negatively impact performance. To "
                    "disable, set `early_stopping=False`.",
                    category=UserWarning)
            if early_stopping is True:
                # Override the early_stopping variable so
                # that it is resolved appropriately in
                # the next block
                early_stopping = "AsyncHyperBandScheduler"
            # Resolve the early stopping object
            early_stopping = resolve_early_stopping(early_stopping, max_iters,
                                                    self._metric_name)

        self.early_stopping = early_stopping
        self.max_iters = max_iters

        self.cv = cv
        self.n_jobs = int(n_jobs or -1)
        self.sk_n_jobs = 1
        if os.environ.get("SKLEARN_N_JOBS") is not None:
            self.sk_n_jobs = int(os.environ.get("SKLEARN_N_JOBS"))

        self.verbose = verbose
        self.error_score = error_score
        self.return_train_score = return_train_score
        self.local_dir = local_dir
        self.use_gpu = use_gpu
        self.loggers = resolve_loggers(loggers)
        assert isinstance(self.n_jobs, int)

    def _fit(self, X, y=None, groups=None, **fit_params):
        """Helper method to run fit procedure

        Args:
            X (:obj:`array-like` (shape = [n_samples, n_features])):
                Training vector, where n_samples is the number of samples and
                n_features is the number of features.
            y (:obj:`array-like`): Shape of array expected to be [n_samples]
                or [n_samples, n_output]). Target relative to X for
                classification or regression; None for unsupervised learning.
            groups (:obj:`array-like` (shape (n_samples,)), optional):
                Group labels for the samples used while splitting the dataset
                into train/test set. Only used in conjunction with a "Group"
                `cv` instance (e.g., `GroupKFold`).
            **fit_params (:obj:`dict` of str): Parameters passed to
                the ``fit`` method of the estimator.

        Returns:
            :obj:`TuneBaseSearchCV` child instance, after fitting.
        """

        self._check_params()
        classifier = is_classifier(self.estimator)
        cv = check_cv(cv=self.cv, y=y, classifier=classifier)
        self.n_splits = cv.get_n_splits(X, y, groups)
        if not hasattr(self, "is_multi"):
            self.scoring, self.is_multi = _check_multimetric_scoring(
                self.estimator, self.scoring)
        else:
            self.scoring, _ = _check_multimetric_scoring(
                self.estimator, self.scoring)

        if self.is_multi:
            if self.refit and (not isinstance(self.refit, str)
                               or self.refit not in self.scoring):
                raise ValueError("When using multimetric scoring, refit "
                                 "must be the name of the scorer used to "
                                 "pick the best parameters. If not needed, "
                                 "set refit to False")

        assert isinstance(
            self.n_jobs,
            int), ("Internal error: self.n_jobs must be an integer.")
        if self.n_jobs < 0:
            resources_per_trial = {"cpu": 1, "gpu": 1 if self.use_gpu else 0}
            if self.n_jobs < -1:
                warnings.warn(
                    "`self.n_jobs` is automatically set "
                    "-1 for any negative values.",
                    category=UserWarning)
        else:
            available_cpus = multiprocessing.cpu_count()
            gpu_fraction = 1 if self.use_gpu else 0
            if ray.is_initialized():
                available_cpus = ray.cluster_resources()["CPU"]
                if self.use_gpu:
                    available_gpus = ray.cluster_resources()["GPU"]
                    gpu_fraction = available_gpus / self.n_jobs
            cpu_fraction = available_cpus / self.n_jobs
            if cpu_fraction > 1:
                cpu_fraction = int(np.ceil(cpu_fraction))
            if gpu_fraction > 1:
                gpu_fraction = int(np.ceil(gpu_fraction))
            resources_per_trial = {"cpu": cpu_fraction, "gpu": gpu_fraction}

        X_id = ray.put(X)
        y_id = ray.put(y)

        config = {}
        config["early_stopping"] = bool(self.early_stopping)
        config["early_stop_type"] = self.early_stop_type
        config["X_id"] = X_id
        config["y_id"] = y_id
        config["groups"] = groups
        config["cv"] = cv
        config["fit_params"] = fit_params
        config["scoring"] = self.scoring
        config["max_iters"] = self.max_iters
        config["return_train_score"] = self.return_train_score
        config["n_jobs"] = self.sk_n_jobs
        config["metric_name"] = self._metric_name

        self._fill_config_hyperparam(config)
        analysis = self._tune_run(config, resources_per_trial)

        self.cv_results_ = self._format_results(self.n_splits, analysis)

        metric = self._metric_name
        base_metric = self._base_metric_name

        # For multi-metric evaluation, store the best_index, best_params and
        # best_score iff refit is one of the scorer names
        # In single metric evaluation, refit_metric is "score"
        if self.refit or not self.is_multi:
            # If callable, refit is expected to return the index of the best
            # parameter set.
            if callable(self.refit):
                self.best_index = self.refit(self.cv_results_)
                if not isinstance(self.best_index, numbers.Integral):
                    raise TypeError("best_index returned is not an integer")
                if (self.best_index < 0
                        or self.best_index >= len(self.cv_results_["params"])):
                    raise IndexError("best_index index out of range")
            else:
                self.best_index = self.cv_results_["rank_test_%s" %
                                                   base_metric].argmin()
                self.best_score = self.cv_results_[
                    "mean_test_%s" % base_metric][self.best_index]
            best_config = analysis.get_best_config(
                metric=metric, mode="max", scope="last")
            self.best_params = self._clean_config_dict(best_config)

        if self.refit:
            base_estimator = clone(self.estimator)
            if self.early_stop_type == EarlyStopping.WARM_START_ENSEMBLE:
                logger.info("tune-sklearn uses `n_estimators` to warm "
                            "start, so this parameter can't be "
                            "set when warm start early stopping. "
                            "`n_estimators` defaults to `max_iters`.")
                if check_is_pipeline(base_estimator):
                    cloned_final_estimator = base_estimator.steps[-1][1]
                    cloned_final_estimator.set_params(
                        **{"n_estimators": self.max_iters})
                else:
                    self.best_params["n_estimators"] = self.max_iters
            # we clone again after setting params in case some
            # of the params are estimators as well.
            self.best_estimator = clone(
                base_estimator.set_params(**self.best_params))
            refit_start_time = time.time()
            if y is not None:
                self.best_estimator.fit(X, y, **fit_params)
            else:
                self.best_estimator.fit(X, **fit_params)
            refit_end_time = time.time()
            self.refit_time = refit_end_time - refit_start_time

        return self

    def fit(self, X, y=None, groups=None, **fit_params):
        """Run fit with all sets of parameters.

        ``tune.run`` is used to perform the fit procedure.

        Args:
            X (:obj:`array-like` (shape = [n_samples, n_features])):
                Training vector, where n_samples is the number of samples and
                n_features is the number of features.
            y (:obj:`array-like`): Shape of array expected to be [n_samples]
                or [n_samples, n_output]). Target relative to X for
                classification or regression; None for unsupervised learning.
            groups (:obj:`array-like` (shape (n_samples,)), optional):
                Group labels for the samples used while splitting the dataset
                into train/test set. Only used in conjunction with a "Group"
                `cv` instance (e.g., `GroupKFold`).
            **fit_params (:obj:`dict` of str): Parameters passed to
                the ``fit`` method of the estimator.

        Returns:
            :obj:`TuneBaseSearchCV` child instance, after fitting.

        """
        ray_init = ray.is_initialized()
        try:
            if not ray_init:
                if self.n_jobs == 1:
                    ray.init(
                        local_mode=True,
                        configure_logging=False,
                        ignore_reinit_error=True,
                        include_dashboard=False)
                else:
                    ray.init(
                        ignore_reinit_error=True,
                        configure_logging=False,
                        include_dashboard=False
                        # log_to_driver=self.verbose == 2
                    )
                    if self.verbose != 2:
                        logger.info("TIP: Hiding process output by default. "
                                    "To show process output, set verbose=2.")

            result = self._fit(X, y, groups, **fit_params)

            if not ray_init and ray.is_initialized():
                ray.shutdown()

            return result

        except Exception:
            if not ray_init and ray.is_initialized():
                ray.shutdown()
            raise

    def score(self, X, y=None):
        """Compute the score(s) of an estimator on a given test set.

        Args:
            X (:obj:`array-like` (shape = [n_samples, n_features])): Input
                data, where n_samples is the number of samples and
                n_features is the number of features.
            y (:obj:`array-like`): Shape of array is expected to be
                [n_samples] or [n_samples, n_output]). Target relative to X
                for classification or regression. You can also pass in
                None for unsupervised learning.

        Returns:
            float: computed score

        """
        self._check_is_fitted(self._metric_name)
        if self.is_multi:
            scorer_name = self.refit
        else:
            scorer_name = "score"
        return self.scoring[scorer_name](self.best_estimator_, X, y)

    def _can_early_stop(self):
        """Helper method to determine if it is possible to do early stopping.

        Only sklearn estimators with `partial_fit` or `warm_start` can be early
        stopped. warm_start works by picking up training from the previous
        call to `fit`.

        Returns:
            bool: if the estimator can early stop

        """
        return self.early_stop_type != EarlyStopping.NO_EARLY_STOP

    def _fill_config_hyperparam(self, config):
        """Fill in the ``config`` dictionary with the hyperparameters.

        For RandomizedSearchCV, samples are pulled from the distribution
        to be saved in the ``config`` dictionary.
        For GridSearchCV, the list is directly saved in the ``config``
        dictionary.

        Implement this functionality in a child class.

        Args:
            config (:obj:`dict`): dictionary to be filled in as the
                configuration for `tune.run`.

        """
        raise NotImplementedError("Define in child class")

    def _tune_run(self, config, resources_per_trial):
        """Wrapper to call ``tune.run``. Implement this in a child class.

        Args:
            config (:obj:`dict`): dictionary to be passed in as the
                configuration for `tune.run`.
            resources_per_trial (:obj:`dict` of int): dictionary specifying the
                number of cpu's and gpu's to use to train the model.

        """
        raise NotImplementedError("Define in child class")

    def _clean_config_dict(self, config):
        """Helper to remove keys from the ``config`` dictionary returned from
        ``tune.run``.

        Args:
            config (:obj:`dict`): Dictionary of all hyperparameter
                configurations and extra output from ``tune.run``., Keys for
                hyperparameters are the hyperparameter variable names
                and the values are the numeric values set to those variables.

        Returns:
            config (:obj:`dict`): Dictionary of all hyperparameter
                configurations without the output from ``tune.run``., Keys for
                hyperparameters are the hyperparameter variable names
                and the values are the numeric values set to those variables.
        """
        for key in [
                "estimator_list", "early_stopping", "X_id", "y_id", "groups",
                "cv", "fit_params", "scoring", "max_iters",
                "return_train_score", "n_jobs", "metric_name",
                "early_stop_type"
        ]:
            config.pop(key, None)
        return config

    def _format_results(self, n_splits, out):
        """Helper to generate the ``cv_results_`` dictionary.

        Args:
            n_splits (int): integer specifying the number of folds when doing
                cross-validation.
            out (:obj:`ExperimentAnalysis`): Object returned by `tune.run`.

        Returns:
            results (:obj:`dict`): Dictionary of results to use for the
                interface's ``cv_results_``.

        """
        trials = [
            trial for trial in out.trials if trial.status == Trial.TERMINATED
        ]
        trial_dirs = [trial.logdir for trial in trials]
        # The result dtaframes are indexed by their trial logdir
        trial_dfs = out.fetch_trial_dataframes()

        # Try to find a template df to use for trials that did not return
        # any results. These trials should copy the structure and fill it
        # with NaNs so that the later reshape actions work.
        template_df = None
        fix_trial_dirs = []  # Holds trial dirs with no results
        for trial_dir in trial_dirs:
            if trial_dir in trial_dfs and template_df is None:
                template_df = trial_dfs[trial_dir]
            elif trial_dir not in trial_dfs:
                fix_trial_dirs.append(trial_dir)

        # Create NaN dataframes for trials without results
        if fix_trial_dirs:
            if template_df is None:
                # No trial returned any results
                return {}
            for trial_dir in fix_trial_dirs:
                trial_df = pd.DataFrame().reindex_like(template_df)
                trial_dfs[trial_dir] = trial_df

        # Keep right order
        dfs = [trial_dfs[trial_dir] for trial_dir in trial_dirs]
        finished = [df.iloc[[-1]] for df in dfs]
        test_scores = {}
        train_scores = {}
        for name in self.scoring:
            test_scores[name] = [
                df[[
                    col for col in dfs[0].columns
                    if "split" in col and "test_%s" % name in col
                ]].to_numpy() for df in finished
            ]
            if self.return_train_score:
                train_scores[name] = [
                    df[[
                        col for col in dfs[0].columns
                        if "split" in col and "train_%s" % name in col
                    ]].to_numpy() for df in finished
                ]
            else:
                train_scores = None

        configs = [trial.config for trial in trials]
        candidate_params = [
            self._clean_config_dict(config) for config in configs
        ]

        results = {"params": candidate_params}
        n_candidates = len(candidate_params)

        def _store(
                results,
                key_name,
                array,
                n_splits,
                n_candidates,
                weights=None,
                splits=False,
                rank=False,
        ):
            """A small helper to store the scores/times to the cv_results_"""
            # When iterated first by n_splits and then by parameters
            array = np.array(
                array, dtype=np.float64).reshape((n_candidates, n_splits))
            if splits:
                for split_i in range(n_splits):
                    results["split%d_%s" % (split_i,
                                            key_name)] = array[:, split_i]

            array_means = np.average(array, axis=1, weights=weights)
            results["mean_%s" % key_name] = array_means
            # Weighted std is not directly available in numpy
            array_stds = np.sqrt(
                np.average(
                    (array - array_means[:, np.newaxis])**2,
                    axis=1,
                    weights=weights))
            results["std_%s" % key_name] = array_stds

            if rank:
                results["rank_%s" % key_name] = np.asarray(
                    rankdata(-array_means, method="min"), dtype=np.int32)

        for name in self.scoring:
            _store(
                results,
                "test_%s" % name,
                test_scores[name],
                n_splits,
                n_candidates,
                splits=True,
                rank=True,
            )
        if self.return_train_score:
            for name in self.scoring:
                _store(
                    results,
                    "train_%s" % name,
                    train_scores[name],
                    n_splits,
                    n_candidates,
                    splits=True,
                    rank=True,
                )

        results["time_total_s"] = np.array(
            [df["time_total_s"].to_numpy() for df in finished]).flatten()

        results["training_iteration"] = np.array([
            df["training_iteration"].to_numpy() for df in finished
        ]).flatten()

        # Use one MaskedArray and mask all the places where the param is not
        # applicable for that candidate. Use defaultdict as each candidate may
        # not contain all the params
        param_results = defaultdict(
            lambda: MaskedArray(
                np.empty(n_candidates),
                mask=True,
                dtype=object,
            )
        )
        for cand_i, params in enumerate(candidate_params):
            for name, value in params.items():
                param_results["param_%s" % name][cand_i] = value

        results.update(param_results)

        return results



import io
import os
from setuptools import setup, find_packages

ROOT_DIR = os.path.dirname(__file__)

# Ray tests depend on buidling tune_sklearn. Thus, if `setup.py`
# depends on Ray, then we create a cyclic dep.
# workaround from: https://stackoverflow.com/a/17626524
with open("tune_sklearn/_version.py") as f:
    text = f.readlines()  # Returns ['__version__ = "0.2.0"']
    __version__ = text[-1].split()[-1].strip("\"'")

VERSION = os.environ.get("TSK_RELEASE_VERSION", __version__)

setup(
    name="tune_sklearn",
    packages=find_packages(),
    version=VERSION,
    author="Michael Chau, Anthony Yu, and Ray Team",
    author_email="ray-dev@googlegroups.com",
    description=("A drop-in replacement for Scikit-Learn’s "
                 "GridSearchCV / RandomizedSearchCV with cutting edge "
                 "hyperparameter tuning techniques."),
    long_description=io.open(
        os.path.join(ROOT_DIR, "README.md"), "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ray-project/tune-sklearn",
    install_requires=["scikit-learn", "scipy", "ray[tune]", "numpy>=1.16"])
